﻿// C# example from Ostrovsky: The C# Memory Model in Theory and
// Practice, in MSDN Magazine, part 2 (January 2013); test harness by
// Jacob Thamsborg; modifications by sestoft@itu.dk 2015-02-27

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

class TestStoreBuffer {
  public static void Main(string[] args) {
    int numberOfAttempts = 1000000;
    int[,] wins = { { 0, 0 }, { 0, 0 } };

    for (var count = 0; count < numberOfAttempts; count++) {
      var sharedState = new StoreBufferExample();
      var taskA = Task.Run(() => { sharedState.ThreadA(); });
      var taskB = Task.Run(() => { sharedState.ThreadB(); });
      Task.WaitAll(taskA, taskB);

      wins[sharedState.A_Won, sharedState.B_Won]++;
    }

    Console.WriteLine("Whew, that took some effort, ({0} tries, sum is {1}) let us recap:", 
		      numberOfAttempts, wins[0,0]+wins[1,0]+wins[0,1]+wins[1,1]);
    Console.WriteLine("No one   won: {0}", wins[0,0]);
    Console.WriteLine("A        won: {0}", wins[1,0]);
    Console.WriteLine("B        won: {0}", wins[0,1]);
    Console.WriteLine("Both (!) won: {0}", wins[1,1]);
  }
}

class StoreBufferExample {
  // On x86 .NET Framework 4.5 and 4.6 and 4.6.2 it makes no
  // difference whether the A and B fields are volatile or not.

  // Moreover Volatile.{Read,Write} does not fix the write order, but
  // two uses of Thread.MemoryBarrier() do.

  public volatile bool A = false, B = false;

  public int A_Won = 0, B_Won = 0;
  
  public void ThreadA() {
    A = true;
    //    Thread.MemoryBarrier();
    if (!B) 
      A_Won = 1;

    // Volatile.Write(ref A, true);
    // if (!Volatile.Read(ref B)) 
    //   A_Won = 1;
  }
  
  public void ThreadB() {
    B = true;
    // Thread.MemoryBarrier();
    if (!A) 
      B_Won = 1;
    
    // Volatile.Write(ref B, true);
    //   if (!Volatile.Read(ref A)) 
    //     B_Won = 1;
  }
}


// Results from .NET 4 on Intel i7 with volatile:
// Z:\cs\pcpp>TestStoreBuffer
// Whew, that took some effort, (1000000 tries, sum is 1000000) let us recap:
// No one   won: 21
// A        won: 930546
// B        won: 56679
// Both (!) won: 12754


// Results from .NET 4 on Intel i7 without volatile:
// Z:\cs\pcpp>TestStoreBuffer
// Whew, that took some effort, (1000000 tries, sum is 1000000) let us recap:
// No one   won: 11
// A        won: 941437
// B        won: 47589
// Both (!) won: 10963

// With Thread.MemoryBarrier() calls, no occurrence of "Both won"
// Whew, that took some effort, (1000000 tries, sum is 1000000) let us recap:
// No one   won: 17359
// A        won: 883306
// B        won: 99335
// Both (!) won: 0

// ----

// Results from Mono 3.12 on Intel i7 with volatile:

// sestoft@mac ~/cs/pcpp $ time mono TestStoreBuffer.exe 
// Whew, that took some effort, (1000000 tries, sum is 1000000) let us recap:
// No one   won: 1
// A        won: 989655
// B        won: 9679
// Both (!) won: 665

// real	0m33.560s
// user	0m34.884s
// sys	0m20.752s

// Results from Mono 3.12 on Intel i7 without volatile:
// sestoft@mac ~/cs/pcpp $ time mono TestStoreBuffer.exe 
// Whew, that took some effort, (1000000 tries, sum is 1000000) let us recap:
// No one   won: 0
// A        won: 991928
// B        won: 7396
// Both (!) won: 676

// real	0m33.569s
// user	0m35.012s
// sys	0m21.031s

// ----

// Results from Mono 4.4.2 on Intel i7 without volatile:
// sestoft@mac ~/cs/pcpp $ time mono TestStoreBuffer.exe 
// Whew, that took some effort, (1000000 tries, sum is 1000000) let us recap:
// No one   won: 0
// A        won: 999962
// B        won: 38
// Both (!) won: 0

// real	1m12.541s
// user	0m57.554s
// sys	1m13.616s


// Results from Mono 4.4.2 on Intel i7 with volatile:
// sestoft@mac ~/cs/pcpp $ time mono TestStoreBuffer.exe 
// Whew, that took some effort, (1000000 tries, sum is 1000000) let us recap:
// No one   won: 0
// A        won: 999958
// B        won: 42
// Both (!) won: 0

// real	1m14.541s
// user	0m57.951s
// sys	1m15.477s
